import csv
import datetime
import queue
import threading
from pathlib import Path
import itertools

import bitalino.cli
from bitalino.device import Bitalino


def write_to_csv(stop_event, data_queue, writer):
    while not stop_event.is_set() or (not data_queue.empty()):
        try:
            # Blocking.
            nSeq, sample, time_stamp = data_queue.get()
            if nSeq % 1000 == 0:
                print(nSeq, *[f"{d:+.2f}" for d in sample], f"{time_stamp:.2f}")
            writer.writerow([nSeq] + list(sample) + [time_stamp])
        except queue.Empty:
            print("No more data.")


def create_csv(data_dir, label):
    start_time = datetime.datetime.now()
    data_dir = Path(data_dir)
    data_dir.mkdir(exist_ok=True)
    csv_path = (
        f"{start_time.isoformat(timespec='seconds')}"
        f"{'' if not label else f'-{label}'}.csv"
    )
    csv_path = data_dir / csv_path
    print(f"Saving data to {csv_path}")
    csvfile = open(csv_path, "w", newline="")
    # Write custom metadata.
    csvfile.write(f"# Start time: {start_time}\n")
    writer = csv.writer(csvfile, dialect="unix", quoting=csv.QUOTE_MINIMAL)
    return csvfile, writer


def main():
    """Write sensor data to a CSV file."""

    parser = bitalino.cli.base_parser("""Stream Bitalino sensor data to a CSV file.""")
    parser.add_argument(
        "--data",
        default="data",
        help="Directory to save CSV.",
    )
    parser.add_argument(
        "--label",
        default="",
        help="Label to append to CSV file name.",
    )
    parser.add_argument(
        "-d",
        "--duration",
        type=int,
        help="Recording duration (s).",
    )

    args = parser.parse_args()

    # Threading
    stop_event = threading.Event()
    data_queue = queue.SimpleQueue()

    # Device setup
    device = Bitalino(args.device)
    device.set_ports(args.ports, args.types)
    device.dumpInfo()
    device.start(
        args.sample_rate,
        data_queue,
        stop_event,
        duration=args.duration,
    )
    device_thread = threading.Thread(target=device.loop)

    # CSV setup
    csvfile, writer = create_csv(args.data, args.label)
    if args.types is None:
        data_cols = args.ports
    else:
        data_cols = [
            p_type if p_type is not None else p
            for p_type, p in itertools.zip_longest(
                device.port_types, device.active_ports
            )
        ]

    writer.writerow(["sample", *data_cols, "time_stamp"])

    data_thread = threading.Thread(
        target=write_to_csv, args=(stop_event, data_queue, writer)
    )

    try:
        data_thread.start()
        device_thread.start()
        device_thread.join()
        data_thread.join()
    except KeyboardInterrupt:
        print("Stopping....")
        stop_event.set()
        device_thread.join()
        data_thread.join()
    finally:
        device.stop()
        device.close()
        print("Device closed")
        csvfile.close()
        print("CSV closed")
