import time

import bitalino.transfer_functions as tf
import plux


class Bitalino(plux.BITalinoDev):
    port_types = {
        "emg": 1,
        "ecg": 2,
        "eda": 3,
        "eeg": 4,
        "acc": 5,
        "lux": 6,
    }

    def __init__(self, address):
        plux.SignalsDev.__init__(f"BTH{address}")
        self.start_time = None
        self.sample_rate = None
        self.duration = None
        self.transfer_functions = None

    def set_ports(self, active_ports=None, port_types=None):
        if active_ports is None:
            active_ports = [self.port_types[p_type] for p_type in port_types]
        if port_types is not None:
            num_types = len(port_types)
            num_ports = len(active_ports)
            if num_types < num_ports:
                port_types = port_types + [None] * (num_ports - num_types)
            elif num_types > num_ports:
                raise ValueError(
                    f"{num_types} port types provided but only {num_ports} ports."
                )
            self.set_transfer_functions(port_types)
        print("Active ports:", active_ports)
        print("Port types:", port_types)
        self.active_ports = active_ports
        self.port_types = port_types

    def onRawFrame(self, nSeq, sample):
        time_stamp = self.get_clock_time()

        # Apply transfer functions.
        if self.transfer_functions is not None:
            sample = [
                f(value) if f is not None else value
                for f, value in zip(self.transfer_functions, sample)
            ]
        self.data_queue.put((nSeq, sample, time_stamp))
        return self.check_stop(nSeq)

    def get_clock_time(self):
        now = time.monotonic()
        if self.start_time is None:
            self.start_time = now
            now = 0.0
        else:
            now = now - self.start_time

        return now

    def start(
        self,
        sample_rate,
        data_queue,
        stop_event,
        duration=None,
    ):
        self.sample_rate = sample_rate
        self.data_queue = data_queue
        self.stop_event = stop_event
        self.duration = duration

        # Third argument is bit-depth but this is ignored by Bitalino devices
        # as it is hardwired to 10-bits.
        super().start(sample_rate, self.active_ports, 0)

    def set_transfer_functions(self, port_types):
        self.transfer_functions = [
            getattr(tf, p_type) if p_type is not None else None for p_type in port_types
        ]

    def check_stop(self, nSeq):
        if self.stop_event.is_set():
            return True
        if self.duration is not None:
            if nSeq > self.duration * self.sample_rate:
                self.stop_event.set()
                return True
        else:
            # Continue
            return False

    def dumpInfo(self):
        print("Device properties:")
        for key, value in self.getProperties().items():
            print(f" - {key}: {value}")
