import multiprocessing as mp
import queue
import threading

from pylsl import StreamOutlet, local_clock
from pylsltools.streams import DataStream

import bitalino.cli
from bitalino.device import Bitalino



class BitalinoLSLDev(Bitalino):

    def get_clock_time(self):
        return local_clock()


class BitalinoStream(DataStream):
    def __init__(
        self,
        name,
        content_type,
        channel_count,
        nominal_srate,
        *,
        source_id="",
        channel_labels=None,
        channel_types=None,
        channel_units=None,
        chunk_size=1,
        max_buffered=360,
        recv_message_queue=None,
        send_message_queue=None,
        debug=False,
        **kwargs,
    ):
        super().__init__(
            name,
            content_type,
            channel_count,
            nominal_srate,
            channel_format="float32",
            source_id=source_id,
            manufacturer="bitalino",
            channel_labels=channel_labels,
            recv_message_queue=recv_message_queue,
            send_message_queue=send_message_queue,
            **kwargs,
        )

        # Initialise local attributes.
        self.chunk_size = chunk_size
        self.max_buffered = max_buffered
        self.debug = debug

    def run(self):
        # Create LSL output
        info = self.make_stream_info()
        outlet = StreamOutlet(info, self.chunk_size, self.max_buffered)

        try:
            while not self.is_stopped():
                # Blocking.
                nSeq, sample, time_stamp = self.recv_message_queue.get(timeout=0.2)
                outlet.push_sample(sample, time_stamp)
                if self.debug and (
                        self.nominal_srate <= 10
                        or (nSeq % self.nominal_srate) == 0
                ):
                    print(self.name, nSeq, *[f"{d:+.2f}" for d in sample], f"{time_stamp:.2f}")
        except queue.Empty:
            print("No data from device.")
        except KeyboardInterrupt:
            print(f"Stopping: {self.name}")
        finally:
            # Call stop on exiting the main loop to ensure cleanup.
            self.stop()
            print(f"Ended: {self.name}.")

    def cleanup(self):
        print(f"Cleanup: {self.name}")


def main():
    """Stream sensor data over LSL."""

    # Use multiprocessing spawn on all platforms.
    mp.set_start_method("spawn")

    parser = bitalino.cli.base_parser("""Stream Bitalino sensor data over LSL.""")
    parser.add_argument(
        "--name",
        default="bitalino",
        help="Stream name.",
    )
    parser.add_argument("--content-type", default="data", help="Stream content type.")
    parser.add_argument(
        "--source-id", default="", help="Unique identifier for stream source."
    )
    parser.add_argument(
        "--chunk-size",
        type=int,
        default=1,
        help="""Desired chunk size in samples. Consumers can override this.""",
    )
    parser.add_argument(
        "--max-buffered",
        type=int,
        default=360,
        help="""Maximum amount of data to buffer - in seconds if there is a
        nominal sampling rate, otherwise x100 in samples.""",
    )
    parser.add_argument(
        "-d",
        "--duration",
        type=int,
        help="Recording duration (s).",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Print extra debugging information."
    )

    args = parser.parse_args()

    if args.ports is not None:
        channel_count = len(args.ports)
    else:
        channel_count = len(args.types)
    nominal_srate = args.sample_rate

    # Threading
    stop_event = threading.Event()

    # Multiprocess: LSL stream runs in its own process.
    data_queue = mp.Queue()

    # Device setup
    device = BitalinoLSLDev(args.device)
    device.set_ports(args.ports, args.types)
    device.dumpInfo()
    device.start(
        args.sample_rate,
        data_queue,
        stop_event,
        duration=args.duration,
    )
    device_thread = threading.Thread(target=device.loop)

    # TODO: Set channel info.
    stream = BitalinoStream(
        args.name,
        args.content_type,
        channel_count,
        nominal_srate,
        source_id=args.source_id,
        channel_labels=None,
        channel_types=None,
        channel_units=None,
        chunk_size=args.chunk_size,
        max_buffered=args.max_buffered,
        recv_message_queue=data_queue,
        send_message_queue=None,
        debug=args.debug,
    )

    try:
        stream.start()
        device_thread.start()
        stream.join()
        # Stop bitalino if LSL stream process returns.
        # FIXME: However, cannot current handle bluetooth disconnections.
        stop_event.set()
        device_thread.join()
    except KeyboardInterrupt:
        print("Stopping....")
        stop_event.set()
        device_thread.join()
        stream.join()
    finally:
        device.stop()
        device.close()
        print("Device closed")
        print("LSL stream closed")
